# external
require 'active_support/cache'
require 'rack/attack'
require 'forwardable'
require 'colorize'

# internal
require 't-1000/version'
require 't-1000/middleware'
require 't-1000/transaction_list'
require 't-1000/transaction'
require 't-1000/request'
require 't-1000/response'
require 't-1000/cache'
require 't-1000/lock'

module T1000
  extend self
  extend Forwardable

  module Refinements
    refine Array do
      def map_regex
        map do |value|
          case value
          when Regexp
            value
          else
            Regexp.new Regexp.escape value.to_s
          end
        end
      end

      def match_any?(string)
        map_regex.any? do |regex|
          string =~ regex
        end
      end

      def stringify
        map(&:to_s).join ', '
      end
    end
  end

  using Refinements

  def_delegators Middleware, *Rack::Attack.methods(false)

  def allow_localhost
    allow_ips '::1', '127.0.0.1'
  end

  def block_ips(*ips, name: nil, &block)
    block ||= proc { true }
    blacklist(humanize __method__, name, *ips) do |req|
      ips.include?(req.ip) && block.call(req)
    end
  end

  def allow_ips(*ips, name: nil, &block)
    block ||= proc { true }
    whitelist(humanize __method__, name, *ips) do |req|
      ips.include?(req.ip) && block.call(req)
    end
  end

  def block_failures(name: nil, within: 30, retries: 10, lock_for: 6000, &block)
    block ||= proc { true }
    blacklist(humanize __method__, name) do |req|
      reqs    = TransactionList.for_ip(req.ip, timeout: within)
      ip_lock = Lock.new(req.ip)
      ip_lock.lock! lock_for if reqs.within(within).with_error.count > retries &&
        block.call(req)
      ip_lock.locked?
    end
  end

  def block_denial_of_service(name: nil, within: 30, requests: 1_000, lock_for: 6000, &block)
    block ||= proc { true }
    blacklist(humanize __method__, name) do |req|
      reqs    = TransactionList.for_ip(req.ip, timeout: within)
      ip_lock = Lock.new(req.ip)
      ip_lock.lock! lock_for if reqs.within(within).count > requests &&
        block.call(req)
      ip_lock.locked?
    end
  end

  def block_strings(*strings, name: nil, &block)
    block ||= proc { true }
    blacklist(humanize __method__, name, *strings) do |req|
      strings.match_any?(req.params.map(&:join).join) && block.call(req)
    end
  end

  def allow_user_agents(*user_agents, name: nil, &block)
    block ||= proc { true }
    whitelist(humanize __method__, name, *user_agents) do
      user_agents.match_any?(req.user_agent) && block.call(req)
    end
  end

  def block_user_agents(*user_agents, name: nil, &block)
    block ||= proc { true }
    blacklist(humanize __method__, name, *user_agents) do
      user_agents.match_any?(req.user_agent) && block.call(req)
    end
  end

  def blacklisted_response(&block)
    Middleware.blacklisted_response = block
  end

  def throttled_response(&block)
    Middleware.throttled_response = block
  end

  def cache
    @cache ||= Cache.new
  end

  def logger=(logger)
    @logger = logger
  end

  def logger
    @logger ||= defined?(Rails) ? Rails.logger : Logger.new(STDOUT)
  end

  def notify(name, type)
    logger.warn case type
    when :whitelist
      "T1000 Allowed Request: #{name.inspect}".green
    when :blacklist
      "T1000 Terminated Request: #{name.inspect}".red
    end
  end

  blacklisted_response do |env|
    notify env['rack.attack.matched'], env['rack.attack.match_type']
    [503, {}, ['503 Service Unavailable']]
  end

  private

  def humanize(meth, name = nil, *vectors)
    "#{meth.to_s.split('_').join(' ').capitalize}#{ " (#{name})" if name }#{ ": #{vectors.stringify}" if vectors.length > 0 }"
  end

end
